from datetime import datetime
from typing import Dict, Any, Union, List
from unittest.mock import patch, MagicMock

import numpy as np
import pytest
import pytz

from arthurai.common.exceptions import UserValueError
from arthurai.util import generate_timestamps, normal_random_ints_fixed_sum

MOCK_NOW_VALUE = datetime(2020, 11, 13, tzinfo=pytz.utc)

generate_timestamp_value_cases = [
    ({'total': 8},
     np.array([1, 1, 1, 1, 1, 1, 1, 1]),
     [pytz.utc.localize(dt) for dt in
      (datetime(2020, 11, 6), datetime(2020, 11, 7), datetime(2020, 11, 8), datetime(2020, 11, 9),
       datetime(2020, 11, 10), datetime(2020, 11, 11), datetime(2020, 11, 12), datetime(2020, 11, 13))]),
    ({'total': 4, 'duration': "2d"},
     np.array([1, 2, 1]),
     [pytz.utc.localize(dt) for dt in
      (datetime(2020, 11, 11), datetime(2020, 11, 12), datetime(2020, 11, 12), datetime(2020, 11, 13))]),
    ({'total': 10, 'duration': "3d"},
     np.array([3, 3, 2, 2]),
     [pytz.utc.localize(dt) for dt in
      (datetime(2020, 11, 10), datetime(2020, 11, 10), datetime(2020, 11, 10), datetime(2020, 11, 11),
       datetime(2020, 11, 11), datetime(2020, 11, 11), datetime(2020, 11, 12), datetime(2020, 11, 12),
       datetime(2020, 11, 13), datetime(2020, 11, 13))]),
    ({'total': 1, 'duration': "3d"}, None, UserValueError),
    ({'total': 11, 'duration': "3h", 'freq': "H"},
     np.array([3, 4, 2, 2]),
     [pytz.utc.localize(dt) for dt in
      (datetime(2020, 11, 12, hour=21), datetime(2020, 11, 12, hour=21), datetime(2020, 11, 12, hour=21),
      datetime(2020, 11, 12, hour=22), datetime(2020, 11, 12, hour=22), datetime(2020, 11, 12, hour=22),
      datetime(2020, 11, 12, hour=22), datetime(2020, 11, 12, hour=23), datetime(2020, 11, 12, hour=23),
      datetime(2020, 11, 13), datetime(2020, 11, 13))]),
    ({'total': 4, 'duration': "2d", 'end': datetime(2020, 12, 13, tzinfo=pytz.utc)},
     np.array([2, 1, 1]),
     [pytz.utc.localize(dt) for dt in
      (datetime(2020, 12, 11), datetime(2020, 12, 11), datetime(2020, 12, 12), datetime(2020, 12, 13))]),
]


@pytest.mark.parametrize("kwargs,repeats,expected", generate_timestamp_value_cases)
def test_generate_timestamps_values(kwargs: Dict[str, Any], repeats: np.array, expected: Union[List[datetime], type]):
    """
    Test that generate_timestamps() outputs match expected values for a few cases
    """
    mock_now = MagicMock()
    mock_now.return_value = MOCK_NOW_VALUE
    mock_randints = MagicMock(normal_random_ints_fixed_sum)
    mock_randints.return_value = repeats
    with patch("arthurai.util._datetime_now", mock_now):
        with patch("arthurai.util.normal_random_ints_fixed_sum", mock_randints):
            if isinstance(expected, list):
                actual = generate_timestamps(**kwargs)
                assert actual == expected
            elif issubclass(expected, Exception):
                with pytest.raises(expected):
                    generate_timestamps(**kwargs)
            else:
                raise TypeError("did not understand 'expected' value")


random_ints_cases = [
    {'num_values': 47, 'total_sum': 58730},
    {'num_values': 47, 'total_sum': 47},
    {'num_values': 13, 'total_sum': 2302},
    {'num_values': 92, 'total_sum': 50238}
]


@pytest.mark.parametrize("kwargs", random_ints_cases)
def test_normal_random_ints_fixed_sum(kwargs: Dict[str, Any]):
    np.random.seed(91)
    actual = normal_random_ints_fixed_sum(**kwargs)
    assert len(actual) == kwargs['num_values']
    assert actual.sum() == kwargs['total_sum']



from arthurai.core.bias.threshold_mitigation import ThresholdMitigation
from arthurai.core.bias.bias_metrics import BiasMetrics


class ArthurBiasWrapper(object):
    """
    This is a wrapper class for all bias-related functionality, including metrics
    and mitigation. This allows users to access bias functionality directly from the
    ArthurModel object: `arthur_model.bias.metrics`, for example. 
    """
    def __init__(self, arthur_model):

        self.model = arthur_model
        self.metrics = BiasMetrics(self.model)
        self.mitigation_threshold = ThresholdMitigation(self.model)


from dataclasses import dataclass
from typing import Any, Dict, Optional
import matplotlib.pyplot as plt
import pandas as pd

from arthurai.common.exceptions import InternalValueError, UserValueError, arthur_excepted, MethodNotApplicableError

DemographicParity= "demographic_parity"
EqualOpportunity = "equal_opportunity"
EqualizedOdds = "equalized_odds"

CONSTRAINT_PLOT_DICT = {
    DemographicParity: ["selection_rate", "accuracy"],
    EqualOpportunity: ["true_positive_rate", "accuracy"],
    EqualizedOdds: ["false_positive_rate", "true_positive_rate"]
}

@dataclass
class Curves(object):
    """
    A class which defines information about a given set of curves for a specific sensitive
    attribute and a specific constraint, as well as a dictionary mapping attribute values 
    to the curves themselves
    """
    attribute_name: str
    constraint: str
    max_acc_idx: int
    attr_val_to_threshold: Dict[str, pd.DataFrame]

class ThresholdMitigation(object):
    def __init__(self, arthur_model): # type not enforced to prevent circular imports, not optimal
        """
        arthur_model should be of type ArthurModel. 
        """
        self.model = arthur_model

    def enable(self):
        """
        Triggers workflow to train curves for all constraints and for all sensitive attributes.
        Equivalent to calling `arthurmodel.enable_bias_mitigation()`; here for convenience.
        """
        return self.model.enable_bias_mitigation()

    @arthur_excepted("Failed to fetch curve for given attribute and constraint")
    def get_curves(self, attribute: str, constraint: str = EqualOpportunity):
        """
        Fetch curves for the given sensitive attribute and constraint.

        :param attribute: Name of the sensitive attribute to fetch curves for. 
        :param constraint: 
        :return: a Curves object which contains information about the attribute, constraint, accuracy-maximizing index, and the actual curves. 
        """

        if constraint not in CONSTRAINT_PLOT_DICT:
            raise UserValueError("`constraint` argument must be one of `demographic_parity`, `equal_opportunity`, or `equalized_odds`.")
        
        attr_id = self.model.get_attribute(attribute).id
        url = f"{self.model._client.base_path}/models/{self.model.id}/enrichments/bias_mitigation/curves?attribute_id={attr_id}&constraint={constraint}"
        resp = self.model._client.get(url)['data'] # returns one dictionary of info per sensitive feature value

        if resp is None:
            raise MethodNotApplicableError("No curves were found at this endpoint; check your attribute and constraint. \
            If you were expecting a curve to be here, something has gone wrong; please reach out to an Arthur engineer.")
        
        res = resp[0]
        best_idx = res['optimization_index']

        curve_result = {}
        for res in resp: 
            if 'categorical_value' in res:
                attr_val = res['categorical_value'] 
            else: 
                attr_val = self._continuous_to_str(res)
            curve_result[attr_val] = pd.DataFrame(res['data_points'])

        return Curves(attribute, constraint, best_idx, curve_result)

    def _continuous_to_str(self, res: Dict[str, Any]):
        """
        For continuous sensitive features only: convert continuous buckets into string categories to use as dict keys.
        """
        start = '(-inf,'
        end = 'inf)'

        if "continuous_start" in res:
            start = '(' + str(res["continuous_start"]) + ','
        if "continuous_end" in res: 
            end = str(res["continuous_end"]) + ']'

        if start == '(-inf,' and end == 'inf)':
            raise InternalValueError("This continuous sensitive attribute only has a single bucket?")

        return start + end

    @arthur_excepted("Failed to plot curves")
    def plot_curves(self, curves: Optional[Curves] = None, attribute: Optional[str] = None, constraint: Optional[str] = None):
        """
        Simple plot of the tradeoff curve for a single sensitive attribute and constraint.

        :param curves: one Curves object containing a set of curves corresponding to a single sensitive attribute and constraint.
        """

        if curves is None:
            if attribute is None or constraint is None:
                raise UserValueError("Either `curves`, or both `attribute` and `constraint`, must be provided.")
            curves = self.get_curves(attribute, constraint)

        ax = plt.axes()

        graph_title = curves.constraint + " " + curves.attribute_name
        x, y = CONSTRAINT_PLOT_DICT[curves.constraint]

        ax.set(title=graph_title, xlabel=x, ylabel=y)

        for sens_val in curves.attr_val_to_threshold:
            points = curves.attr_val_to_threshold[sens_val]
            ax.plot(points['x'], points['y'], ls='-', lw=2.0, label='sensitive feature value = ' + sens_val)
        
        if curves.constraint != EqualizedOdds:
            xbest = list(curves.attr_val_to_threshold.values())[0].iloc[curves.max_acc_idx]['x']
            ax.axvline(x=xbest, label="solution")

        ax.legend()
        plt.show()

    @arthur_excepted("Failed to get thresholds for index")
    def get_thresholds_for_idx(self, curves: Curves, idx: int = None):
        """
        Retrieve the thresholds for each group, given a set of curves and aparticular optimization index. 

        :param curves: one Curves object containing a set of curves corresponding to a single sensitive attribute and constraint.
        :param idx: one integer (between 0 and 1000, inclusive) that represents a specific solution option. Defaults to the curve's accuracy-maximizing index.
        :return: a dictionary mapping sensitive feature values to the prediction threshold used for that group. 
        """

        idx = curves.max_acc_idx if idx is None else idx
        if idx < 0 or idx > 1000:
            raise UserValueError("optimization index should be between 0 and 1000, inclusive.")
        thresholds = {}
        for attr in curves.attr_val_to_threshold:
            thresholds[attr] = curves.attr_val_to_threshold[attr].iloc[idx]['threshold']

        return thresholds



import pytest
from arthurai.core.models import ArthurModel, ExplainabilityParameters
from arthurai.core.attributes import ArthurAttribute
from tests.test_request_models.fixtures import explainability_response_objects, explainability_json_strings, \
    model_attribute_objects, model_attribute_json_strings, model_response_json_strings, model_response_objects, \
    explainability_dicts, model_attribute_dicts, model_response_dicts


class TestExplainabilityParameterObject:
    @pytest.mark.parametrize('explainability_response, explainability_json_string',
                             tuple(zip(explainability_response_objects.values(), explainability_json_strings.values())))
    def test_explainability_response_serialization(self, explainability_response, explainability_json_string):
        # we will skip the 3rd string as we only want to test deserialization of unknown fields
        if explainability_json_string != explainability_json_strings[3]:
            assert explainability_json_string == explainability_response.to_json()

    @pytest.mark.parametrize('explainability_response, explainability_json_string',
                             tuple(zip(explainability_response_objects.values(), explainability_json_strings.values())))
    def test_explainability_response_deserialization(self, explainability_response, explainability_json_string):
        deserialzed_explainability_response = ExplainabilityParameters.from_json(explainability_json_string)
        assert deserialzed_explainability_response == explainability_response

    @pytest.mark.parametrize('explainability_response, explainability_dict',
                             tuple(zip(explainability_response_objects.values(), explainability_dicts.values())))
    def test_explainability_response_to_dict(self, explainability_response, explainability_dict):
        actual_explainability_response_dict = explainability_response.to_dict()
        assert actual_explainability_response_dict == explainability_dict


class TestArthurAttributeObject:
    @pytest.mark.parametrize('model_attribute_response, model_attribute_json_string',
                             tuple(zip(model_attribute_objects.values(), model_attribute_json_strings.values())))
    def test_model_attribute_response_serialization(self, model_attribute_response, model_attribute_json_string):
        # we will skip the 4th string as we only want to test deserialization of unknown fields
        if model_attribute_json_string != model_attribute_json_strings[4]:
            assert model_attribute_json_string == model_attribute_response.to_json()

    @pytest.mark.parametrize('model_attribute_response, model_attribute_json_string',
                             tuple(zip(model_attribute_objects.values(), model_attribute_json_strings.values())))
    def test_model_attribute_response_deserialization_json(self, model_attribute_response, model_attribute_json_string):
        deserialized_model_attribute_response = ArthurAttribute.from_json(model_attribute_json_string)
        assert deserialized_model_attribute_response == model_attribute_response

    @pytest.mark.parametrize('model_attribute_response, model_attribute_dict',
                             tuple(zip(model_attribute_objects.values(), model_attribute_dicts.values())))
    def test_model_attribute_response_deserialization_dict(self, model_attribute_response, model_attribute_dict):
        actual_model_attribute_response_dict = model_attribute_response.to_dict()
        print(actual_model_attribute_response_dict)
        assert actual_model_attribute_response_dict == model_attribute_dict


class TestArthurModelObject:
    @pytest.mark.parametrize('model_response, model_response_json_string',
                             tuple(zip(model_response_objects.values(), model_response_json_strings.values())))
    def test_model_response_serialization(self, model_response, model_response_json_string):
        # we will skip the 3rd string as we only want to test deserialization of unknown fields
        if model_response_json_string != model_response_json_strings[3]:
            assert model_response_json_string == model_response.to_json()
    @pytest.mark.parametrize('model_response, model_response_json_string',
                             tuple(zip(model_response_objects.values(), model_response_json_strings.values())))
    def test_model_response_deserialization(self, model_response, model_response_json_string):
        deserialzed_model_response = ArthurModel.from_json(model_response_json_string)
        assert deserialzed_model_response == model_response
    @pytest.mark.parametrize('model_response, model_response_dict',
                             tuple(zip(model_response_objects.values(), model_response_dicts.values())))
    def test_model_attribute_response_deserialization(self, model_response, model_response_dict):
        actual_model_response_dict = model_response.to_dict()
        assert actual_model_response_dict == model_response_dict
